//Copyright Magnus Hellman 2000

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <OS.h>
#include <scheduler.h>

char* FormatLongLong(long long lVal, char *str)
{
	sprintf(str, "%Ld", lVal);
	if (lVal >= 10000)
	{
		int len = strlen(str);
		int numCommas = (len - 1) / 3;
		int firstCommaPos = len - numCommas * 3;
		for (int comma = numCommas; comma > 0; --comma)
		{
			int thisCommaPos = firstCommaPos + 4 * (comma - 1);
			memmove(str + thisCommaPos + 1,
					str + firstCommaPos + (comma - 1) * 3,
					3);
			str[thisCommaPos] = ',';
		}
		str[len + numCommas] = 0;
	}
	
	return str;
}

int main(int argc, char *argv[])
{
	const char *states[] = {	"invalid thread state!", "running", "ready to run",
								"receiving", "asleep", "suspended",
								"waiting for semaphore" };
	int32 biscuit = 0;
	status_t st;
	thread_info info;
	if (argc == 3 && strcmp(argv[1], "-T") == 0) {
		while ((st = get_next_thread_info(atoi(argv[2]), &biscuit, &info)) == B_OK) {
			char sem[10];
			if (info.state == B_THREAD_WAITING)
				sprintf(sem, " %u", info.sem); //huh? what happened to itoa()?
			else
				*sem = 0;
			if (info.state < 1 || info.state >= sizeof(states) / sizeof(states[0]))
				info.state = (thread_state)0;
				
			printf("%5u: %-*s %s%s\n",
						info.thread, B_OS_NAME_LENGTH, info.name, states[info.state], sem);
		}
		
		if (st != B_BAD_VALUE || biscuit == 0)
			printf("get_next_thread_info(%s, ...) error: 0x%x %s\n", argv[2], st, strerror(st));
	} else
	if (argc >= 2) {
		for (int i = 1; i < argc; ++i) {
			st = get_thread_info(atoi(argv[i]), &info);
			if (i > 1)
				printf("\n");
			if (st == B_OK) {
				char sem[20];
				if (info.state == B_THREAD_WAITING)
					sprintf(sem, " %u", info.sem);
				else
					*sem = 0;
				if (info.state < 1 || info.state >= sizeof(states) / sizeof(states[0]))
					info.state = (thread_state)0;
				char timeStr[2][20];	
				
				printf(	"Thread id:    %s\n"
						"Team id:      %u\n"
						"Name:         %s\n"
						"State:        %s%s\n"
						"Priority:     %u\n"
						"User time:    %s microseconds\n"
						"Kernel time:  %s microseconds\n"
						"Stack base:   %p\n"
						"Stack end:    %p\n",
					argv[i], info.team, info.name, states[info.state], sem, info.priority,
					FormatLongLong(info.user_time, timeStr[0]),
					FormatLongLong(info.kernel_time, timeStr[1]),
					info.stack_base, info.stack_end);
					
				char strTime[21];
				printf("Estimated maximum scheduling latency: %s microseconds\n",
							FormatLongLong(estimate_max_scheduling_latency(info.thread), strTime));
				int32 teamBikkie = 0;
				team_info teamInfo;
				bool firstOne = true;
				printf("This thread has locked the following semaphores (and probably more):\n");
				while (get_next_team_info(&teamBikkie, &teamInfo) == B_OK) {
					int32 semBikkie = 0;
					sem_info semInfo;
					while (get_next_sem_info(teamInfo.team, &semBikkie, &semInfo) == B_OK)
						if (semInfo.latest_holder == info.thread && semInfo.count <= 0) {
							if (firstOne) {
								firstOne = false;
							}
							printf("%6u: %s (team %u, count %d)\n",
									semInfo.sem, semInfo.name, teamInfo.team, semInfo.count);
						}
				}
				if (firstOne)
					printf("  None found\n");
			} else
				printf("get_thread_info(%s) error: 0x%x %s\n", argv[i], st, strerror(st));
		}
	} else
		printf("Usage: \"threadinfo -T <team_id>\" or \"threadinfo <thread_id> ...\"\n");

	return 0;
}
